@@ -1,32 +1,45 @@
 # -*- coding: utf-8 -*-
 # Based heavily on the implementation of pint's Quantity object
 from __future__ import annotations
+
+import copy
+import locale
+import operator
+import warnings
 from re import match
 from types import MethodType
 
-import numpy as np
-import locale
-import copy
-import operator
-import warnings
 import jax
-from pint import Quantity, DimensionalityError
+import numpy as np
+from pint import DimensionalityError, Quantity
 from pint.util import SharedRegistryObject
 
-from .wrap_numpy import wrap_numpy, HANDLED_FUNCTIONS, HANDLED_UFUNCS
 from . import NegativeStdDevError, NumpyDowncastWarning
-from .util import is_np_duck_array, ignore_runtime_warnings, ignore_numpy_downcast_warnings, Display
+from .util import (
+    Display,
+    ignore_numpy_downcast_warnings,
+    ignore_runtime_warnings,
+    is_np_duck_array,
+    strip_device_array,
+)
+from .wrap_numpy import HANDLED_FUNCTIONS, HANDLED_UFUNCS, wrap_numpy
 
 
 def _check_units(value, err):
     mag_units = hasattr(value, "units")
     err_units = hasattr(err, "units")
     if mag_units ^ err_units and err is not None:
-        raise ValueError("Both mag and err need to have units if one of them has units!")
+        raise DimensionalityError(
+            "",
+            "",
+            extra_msg="...Both mag and err need to have units if one of them has units!",
+        )
     if mag_units and err is not None:
         if value.units != err.units:
-            raise ValueError(
-                f"Value units {value.units} cannot be converted to error units {err.units}"
+            raise DimensionalityError(
+                value.units,
+                err.units,
+                extra_msg=f"Value units {value.units} cannot be converted to error units {err.units}",
             )
     if mag_units:
         mag_units = value.units
@@ -36,23 +49,22 @@ def _check_units(value, err):
         else:
             ret_err = None
     else:
-        mag_units = 1.0
+        mag_units = None
         ret_val = value
         ret_err = err
 
     return ret_val, ret_err, mag_units
 
 
-def _strip_device_array(value, err):
-    if isinstance(value, jax.xla.DeviceArray):
-        value = value.to_py().copy()
-    if isinstance(err, jax.xla.DeviceArray):
-        err = err.to_py().copy()
-    return value, err
-
-
 class Uncertainty(Display):
-    __apply_to_both_ndarray__ = ["flatten", "real", "imag", "astype", "T"]
+    __apply_to_both_ndarray__ = [
+        "flatten",
+        "real",
+        "imag",
+        "astype",
+        "T",
+        "reshape",
+    ]
     __ndarray_attributes__ = ["dtype", "ndim", "size"]
 
     # Pint comparibility
@@ -79,13 +91,17 @@ class Uncertainty(Display):
             Dimensionality of the Quantity, e.g. ``{length: 1, time: -1}``
         """
         if self._dimensionality is None:
-            self._dimensionality = self._REGISTRY._get_dimensionality(self.units)
+            self._dimensionality = self._REGISTRY._get_dimensionality(
+                self.units
+            )
 
         return self._dimensionality
 
     def check(self, dimension) -> bool:
         """Return true if the quantity's dimension matches passed dimension."""
-        return self.dimensionality == self._REGISTRY.get_dimensionality(dimension)
+        return self.dimensionality == self._REGISTRY.get_dimensionality(
+            dimension
+        )
 
     def _check(self, other) -> bool:
         """Check if the other object use a registry and if so that it is the
@@ -104,7 +120,9 @@ class Uncertainty(Display):
 
         elif isinstance(other, SharedRegistryObject):
             mess = "Cannot operate with {} and {} of different registries."
-            raise ValueError(mess.format(self.__class__.__name__, other.__class__.__name__))
+            raise ValueError(
+                mess.format(self.__class__.__name__, other.__class__.__name__)
+            )
         else:
             return False
 
@@ -142,41 +160,46 @@ class Uncertainty(Display):
     def _convert_magnitude_not_inplace(self, other, *contexts, **ctx_kwargs):
         if contexts:
             with self._REGISTRY.context(*contexts, **ctx_kwargs):
-                return self._REGISTRY.convert(self._magnitude, self._units, other)
+                return self._REGISTRY.convert(
+                    self._magnitude, self._units, other
+                )
 
         return self._REGISTRY.convert(self._magnitude, self._units, other)
 
     @ignore_numpy_downcast_warnings
     def __init__(self, value, err=None):
 
-        value, err, units = _check_units(value, err)
-        value, err = _strip_device_array(value, err)
+        value_, err_, units = _check_units(value, err)
+        value_ = strip_device_array(value_)
+        err_ = strip_device_array(err_)
 
         # If Uncertatity
-        if isinstance(value, self.__class__):
-            magnitude_nom = value.value
-            magnitude_err = value.error
+        if isinstance(value_, self.__class__):
+            magnitude_nom = value_.value
+            magnitude_err = value_.error
         # If sequence
-        elif isinstance(value, list):
-            return self.__class__.from_list(value)
+        elif isinstance(value_, list):
+            inst = self.__class__.from_list(value_)
+            magnitude_nom = inst.value
+            magnitude_err = inst.error
         # If arrays
-        elif np.ndim(value) > 0:
-            magnitude_nom = np.asarray(value)
-            if err is None:
-                magnitude_err = np.zeros_like(value)
+        elif np.ndim(value_) > 0:
+            magnitude_nom = np.asarray(value_)
+            if err_ is None:
+                magnitude_err = np.zeros_like(value_)
             else:
-                if np.ndim(err) == 0:
-                    magnitude_err = np.ones_like(value) * err
+                if np.ndim(err_) == 0:
+                    magnitude_err = np.ones_like(value_) * err_
                 else:
-                    magnitude_err = np.asarray(err)
+                    magnitude_err = np.asarray(err_)
                     assert magnitude_err.shape == magnitude_nom.shape
         # If scalar
         else:
-            magnitude_nom = value
-            if err is None:
+            magnitude_nom = value_
+            if err_ is None:
                 magnitude_err = 0.0
             else:
-                magnitude_err = err
+                magnitude_err = err_
 
         # Replace NaNs in errors with zeros
         if is_np_duck_array(type(magnitude_err)):
@@ -184,8 +207,10 @@ class Uncertainty(Display):
         else:
             if not np.isfinite(magnitude_err):
                 magnitude_err = 0
-        magnitude_nom *= units
-        magnitude_err *= units
+
+        if units is not None:
+            magnitude_nom *= units
+            magnitude_err *= units
         # Basic sanity checks
         if is_np_duck_array(type(magnitude_nom)):
             match_items = self.__ndarray_attributes__ + ["shape"]
@@ -194,7 +219,9 @@ class Uncertainty(Display):
             except ValueError:
                 pass
             for item in match_items:
-                if not getattr(magnitude_nom, item) == getattr(magnitude_err, item):
+                if not getattr(magnitude_nom, item) == getattr(
+                    magnitude_err, item
+                ):
                     raise ValueError(
                         f"Attribute {item} does not match for value and error! ({getattr(magnitude_nom,item)} and {getattr(magnitude_err,item)})"
                     )
@@ -222,7 +249,9 @@ class Uncertainty(Display):
         return ret
 
     def __deepcopy__(self, memo) -> Uncertainty:
-        ret = self.__class__(copy.deepcopy(self._nom, memo), copy.deepcopy(self._err, memo))
+        ret = self.__class__(
+            copy.deepcopy(self._nom, memo), copy.deepcopy(self._err, memo)
+        )
         return ret
 
     def __hash__(self) -> int:
@@ -255,9 +284,11 @@ class Uncertainty(Display):
     @property
     def relative(self):
         if np.ndim(self._nom) == 0:
-            if self._nom != 0:
+            try:
                 return self._err / self._nom
-            else:
+            except (OverflowError):
+                return np.inf
+            except ZeroDivisionError:
                 return np.NaN
         else:
             return self._err / self._nom
@@ -266,6 +297,13 @@ class Uncertainty(Display):
     def rel(self):
         return self.relative
 
+    @property
+    def rel2(self):
+        try:
+            return self.relative**2
+        except OverflowError:
+            return np.inf
+
     @classmethod
     def from_list(cls, u_list):
         return cls.from_sequence(u_list)
@@ -320,7 +358,7 @@ class Uncertainty(Display):
     def __add__(self, other):
         if isinstance(other, Uncertainty):
             new_mag = self._nom + other._nom
-            new_err = np.sqrt(self._err ** 2 + other._err ** 2)
+            new_err = np.sqrt(self._err**2 + other._err**2)
         else:
             new_mag = self._nom + other
             new_err = self._err
@@ -340,7 +378,7 @@ class Uncertainty(Display):
     def __sub__(self, other):
         if isinstance(other, Uncertainty):
             new_mag = self._nom - other._nom
-            new_err = np.sqrt(self._err ** 2 + other._err ** 2)
+            new_err = np.sqrt(self._err**2 + other._err**2)
         else:
             new_mag = self._nom - other
             new_err = self._err
@@ -361,7 +399,7 @@ class Uncertainty(Display):
     def __mul__(self, other):
         if isinstance(other, Uncertainty):
             new_mag = self._nom * other._nom
-            new_err = np.abs(new_mag) * np.sqrt(self.rel ** 2 + other.rel ** 2)
+            new_err = np.abs(new_mag) * np.sqrt(self.rel2 + other.rel2)
         else:
             new_mag = self._nom * other
             new_err = np.abs(self._err * other)
@@ -385,7 +423,7 @@ class Uncertainty(Display):
         # Self / Other
         if isinstance(other, Uncertainty):
             new_mag = self._nom / other._nom
-            new_err = np.abs(new_mag) * np.sqrt(self.rel ** 2 + other.rel ** 2)
+            new_err = np.abs(new_mag) * np.sqrt(self.rel2 + other.rel2)
         else:
             new_mag = self._nom / other
             new_err = np.abs(self._err / other)
@@ -466,7 +504,7 @@ class Uncertainty(Display):
         return other // self, other % self
 
     def __ipow__(self, other):
-        new = self ** other
+        new = self**other
         if is_np_duck_array(type(self._nom)):
             self._err = new._err
             self._nom = new._nom
@@ -485,11 +523,8 @@ class Uncertainty(Display):
         else:
             B = other
             sB = 0
-        new_mag = A ** B
-        if sB == 0 and int(B) == B:
-            new_err = np.abs(new_mag) * np.sqrt((B / A * sA) ** 2)
-        else:
-            new_err = np.sqrt((B / A * sA) ** 2 + (np.log(A) * sB) ** 2)
+        new_mag = A**B
+        new_err = np.abs(new_mag) * np.sqrt((B / A * sA) ** 2)
 
         return self.__class__(new_mag, new_err)
 
@@ -505,8 +540,10 @@ class Uncertainty(Display):
             A = other
             sA = 0
 
-        new_mag = A ** B
-        new_err = np.abs(new_mag) * np.sqrt((B / A * sA) ** 2 + (np.log(A) * sB) ** 2)
+        new_mag = A**B
+        new_err = np.abs(new_mag) * np.sqrt(
+            (B / A * sA) ** 2 + (np.log(A) * sB) ** 2
+        )
 
         return self.__class__(new_mag, new_err)
 
@@ -572,14 +609,18 @@ class Uncertainty(Display):
             raise NotImplementedError
         else:
             if ufunc.__name__ not in HANDLED_UFUNCS:
-                raise NotImplementedError(f"Ufunc {ufunc.__name__} is not implemented!") from None
+                raise NotImplementedError(
+                    f"Ufunc {ufunc.__name__} is not implemented!"
+                ) from None
             else:
                 return wrap_numpy("ufunc", ufunc, args, kwargs)
 
     def __getattr__(self, item):
         if item.startswith("__array_"):
             # Handle array protocol attributes other than `__array__`
-            raise AttributeError(f"Array protocol attribute {item} not available.")
+            raise AttributeError(
+                f"Array protocol attribute {item} not available."
+            )
         elif item in self.__apply_to_both_ndarray__:
             val = getattr(self._nom, item)
             err = getattr(self._err, item)
@@ -591,9 +632,13 @@ class Uncertainty(Display):
             else:
                 return self.__class__(val, err)
         elif item in HANDLED_UFUNCS:
-            return lambda *args, **kwargs: wrap_numpy("ufunc", item, [self] + list(args), kwargs)
+            return lambda *args, **kwargs: wrap_numpy(
+                "ufunc", item, [self] + list(args), kwargs
+            )
         elif item in HANDLED_FUNCTIONS:
-            return lambda *args, **kwargs: wrap_numpy("function", item, [self] + list(args), kwargs)
+            return lambda *args, **kwargs: wrap_numpy(
+                "function", item, [self] + list(args), kwargs
+            )
         elif item in self.__ndarray_attributes__:
             return getattr(self._nom, item)
         elif hasattr(Quantity, item):
@@ -636,7 +681,9 @@ class Uncertainty(Display):
 
     def clip(self, min=None, max=None, out=None, **kwargs):
 
-        return self.__class__(self._nom.clip(min, max, out, **kwargs), self._err)
+        return self.__class__(
+            self._nom.clip(min, max, out, **kwargs), self._err
+        )
 
     def fill(self, value) -> None:
         return self._nom.fill(value)
@@ -646,7 +693,9 @@ class Uncertainty(Display):
             self._nom.put(indices, values._nom, mode)
             self._err.put(indices, values._err, mode)
         else:
-            raise ValueError("Can only 'put' Uncertainties into uncertainties!")
+            raise ValueError(
+                "Can only 'put' Uncertainties into uncertainties!"
+            )
 
     def copy(self):
         return Uncertainty(self._nom.copy(), self._err.copy())
@@ -690,7 +739,9 @@ class Uncertainty(Display):
         try:
             _ = self._nom[key]
         except ValueError as exc:
-            raise ValueError(f"Object {type(self._nom)} does not support indexing") from exc
+            raise ValueError(
+                f"Object {type(self._nom)} does not support indexing"
+            ) from exc
 
         self._nom[key] = value._nom
         self._err[key] = value._err
@@ -703,8 +754,16 @@ class Uncertainty(Display):
                 return self.__class__(nom, err)
             else:
                 return [
-                    self.__class__(n, e).tolist() if isinstance(n, list) else self.__class__(n, e)
+                    self.__class__(n, e).tolist()
+                    if isinstance(n, list)
+                    else self.__class__(n, e)
                     for n, e in (nom, err)
                 ]
         except AttributeError:
-            raise AttributeError(f"{type(self._nom).__name__}' does not support tolist.")
+            raise AttributeError(
+                f"{type(self._nom).__name__}' does not support tolist."
+            )
+
+    @property
+    def ndim(self):
+        return np.ndim(self._nom)
